/* Sample code Donutstorm 1 by Rudy Rucker, CS Dept, SJSU, Sept 25, 2003.
This code is public domain and can be freely reused. */

#include <windows.h>   // use as needed for your system
#include <gl\gl.h>  //Must include this before vectormatrix
#include <gl\glut.h>
#include <math.h> //For cos and sine functions
#include "timer.h"
#include "randomizer.h"
#include "vectormatrix.h"
#include "polyhedron.h"
#include <deque> //For "deque" template used for deques, which are double-ended queues
using namespace std; //Need this for standard template library to work.

cPerformanceTimer timer; //A class I defined in timer.h

GLdouble ZOOMRATIO = 0.8;
GLdouble startwindowwidth = 640.0;
GLdouble aspect = 4.0/3.0;
GLdouble xradius = 20.0;
GLdouble yradius = 15.0;
GLdouble zradius = 50.0;
cVector eye(0.0, - 0.05 * yradius, - 1.1 * zradius);
double zoom = 0.8;
GLdouble MINTIMEPERFRAME = (1.0/60.0); //Min time in seconds per frame
Real dt;
bool animateflag = true;
bool threedflag = true;
bool lightingflag = true; 
int POLYHEDRONCOUNT = 35; //Number of polyhedra to have in the storm.
Real FALLSPEED = 2.0*zradius / 8.0; /* Means it should take 8.0 seconds
	to fall from hiz to loz. */

#define BLACKBACKGROUND
#define ANTIALIAS
#define NUMBEROFLIGHTS 4 
/* You can try 1, 2, 3, or 4 to see how it changes the look. */

void myLighting(bool onoff); //Body of this function is defined below.

//<<<<<<<<<<<<<<<<<<<<<<< myInit >>>>>>>>>>>>>>>>>>>>

void setProjectionMatrix(double zoom)
{
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
	if (!threedflag) //Use a 2D ortho projection.
		glOrtho(-zoom *xradius, zoom *xradius, -zoom *yradius, zoom *yradius,
			 -zradius, zradius); 
		/* Arguments: left, right, bottom, top, near, far.  The near and far are
		relative to a viewer located at the origin, so here we normally
		want a negative near and a positive far. We zoom only in the z and y,
		as zooming in the z might clip unnecessariy. */
	else //Use a 3D perspective projection
	{
		cVector hicorner(xradius, yradius, zradius);
		double eyetohicorner = (hicorner - eye).magnitude();
		gluPerspective(zoom*45.0, aspect, 0.1, eyetohicorner); //eyetohicorner);
		/* Arguments: view angle in degrees, width-to-height aspect ratio, 
		near, far.  The aspect should match the viewport dx/dy.  The near
		and far should be both positive, as here we think of the eye as
		standing back from the scene. */
		gluLookAt(eye.x(), eye.y(), eye.z(), 0.0, 0.0, 0.0, 0.0, 0.0, 1.0);
		/* Arguments: 3 eye coords, 3 center coords, 3 up direction components.
		This puts the eye at the orgin of a viewer coordinate system and puts
		the "center" on the negative z axis. */
	}
}

void myInit(void)
{
/* We call myInit once at startup. */
#ifdef BLACKBACKGROUND
    glClearColor(0.0, 0.0, 0.0, 0.0);       // set black background color
#else //not BLACKBACKGROUND
    glClearColor(1.0, 1.0, 1.0, 0.0);       // set white background color
#endif //BLACKBACKGROUND
	setProjectionMatrix(1.0);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
	glShadeModel(GL_FLAT);
	glEnable(GL_DEPTH_TEST); //In case I want to use 3D
	myLighting(lightingflag);
	timer.setMinDt(MINTIMEPERFRAME); //Don't run faster than this rate
}

//<<<<<<<<<<<<<<<<<<<<<<< drawFlurry >>>>>>>>>>>>>>>>>
void drawFlurry() //Uses the global dt param.
{
	static deque<cPolyhedron*> polys; //Seems faster and cleaner to use pointers.
	cPolyhedron *pnewpoly;
	while (polys.size() < POLYHEDRONCOUNT)
	{
		pnewpoly = new cPolyhedron(); //Remember to delete it when you're done with it
		pnewpoly->randomize(); 
		pnewpoly->setPosition(cVector(cRandomizer::pinst()->randomReal(-0.5* xradius, 0.5 * xradius),
			cRandomizer::pinst()->randomReal(-0.5* yradius, 0.5* yradius),
			cRandomizer::pinst()->randomReal(-zradius, zradius)));
		polys.push_back(pnewpoly); //push_back adds a copy of the polygon to the array.
	}
	for (int i=0; i<polys.size(); i++)
	{
		polys[i]->animateTumble(dt);
		polys[i]->translate(-FALLSPEED * dt * polys[i]->radius() * cVector::ZAXIS);
		polys[i]->draw();
		if (polys[i]->position().z() < -zradius) 
		{
			delete polys[i]; //Delete it before removing from the deque.
			polys.erase(polys.begin()+i); //Erase when you hit bottom.
		}
	}
}  
//<<<<<<<<<<<<<<<<<<<<<<<< myDisplay >>>>>>>>>>>>>>>>>
void myDisplay(void)
{
/* myDisplay is called once at start up and is called again whenever
the graphics window is first covered and then uncovered, making a
redraw necessary. */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);     // clear the screen 
//Draw 3D box around the world.
	setMaterialColor(cVector(0.8, 0.8, 0.8)); //Light gray
	glLineWidth(2.0);
	glBegin(GL_LINE_LOOP); //loz face
		glVertex3d(-xradius, -yradius, -zradius);
		glVertex3d(xradius, -yradius, -zradius);
		glVertex3d(xradius, yradius, -zradius);
		glVertex3d(-xradius, yradius, -zradius);
	glEnd();
	glBegin(GL_LINE_LOOP); //hizface
		glVertex3d(-xradius, -yradius, zradius);
		glVertex3d(xradius, -yradius, zradius);
		glVertex3d(xradius, yradius, zradius);
		glVertex3d(-xradius, yradius, zradius);
	glEnd();
	glBegin(GL_LINE_LOOP); //loyface
		glVertex3d(-xradius, -yradius, zradius);
		glVertex3d(xradius, -yradius, zradius);
		glVertex3d(xradius, -yradius, -zradius);
		glVertex3d(-xradius, -yradius, -zradius);
	glEnd();
	glBegin(GL_LINE_LOOP); //hiyface
		glVertex3d(-xradius, yradius, zradius);
		glVertex3d(xradius, yradius, zradius);
		glVertex3d(xradius, yradius, -zradius);
		glVertex3d(-xradius, yradius, -zradius);
	glEnd();
	//Draw coordinate axes in Red, Blue, Yellow
	glLineWidth(3.0);
	glBegin(GL_LINES);
		setMaterialColor(cVector(1.0, 0.0, 0.0)); //X axis red
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(1.0, 0.0, 0.0);
		setMaterialColor(cVector(0.0, 0.0, 1.0));//Y axis blue
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(0.0, 1.0, 0.0);
		setMaterialColor(cVector(1.0, 1.0, 0.0)); //Z axis yellow
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(0.0, 0.0, 1.0);
	glEnd();
	//Draw polygons
	drawFlurry();
	glFlush();				// Single buffer: send all output to display 
	glutSwapBuffers();		//Double buffer: flip.
}
//<<<<<<<<<<<<<<<<<<<<<<<< myReshape >>>>>>>>>>>>>>>>>

void myReshape(int width, int height)
{
	aspect = double(width)/double(height);
	setProjectionMatrix(zoom);
	glViewport(0, 0, width, height);
}
//<<<<<<<<<<<<<<<<<<<<<<<< myAnimate >>>>>>>>>>>>>>>>>
void myAnimate(void)
{
/* myAnimate is called over and over whenever glut is idle (not busy
processing a key or mouse message). To keep this from happening too
often we put in a delay, that is, we use a spin loop to make myAnimate
take at least MINTIMEPERFRAME seconds. */
	if (!animateflag)
		return;
	dt = timer.tick();
	myDisplay();
}
//<<<<<<<<<<<<<<<<<<<<<<<< myKeyboard >>>>>>>>>>>>>>>>>
void myKeyboard(unsigned char key, int mx, int my)
{
/* key is the ASCII code of the key, such as 'a', 'A', 'b', 'B', and
so on.  Note that case matters.  mx and my are the current mouse
position. */
	switch(key)
	{
		case ' ': //Space
			myDisplay(); //redraw 
			break;
		case 'a':
			animateflag ^= TRUE;
			break;
		case 'q':
			exit(-1); //Terminate
			break;
		case 'r':
			break;
		case 'z':
			zoom *= ZOOMRATIO;
			setProjectionMatrix(zoom);
			break;
		case 'Z':
			zoom *= 1.0/ZOOMRATIO;
			setProjectionMatrix(zoom);
			break;
		case '2':
			threedflag = false;
			setProjectionMatrix(zoom);
			break;
		case '3':
			threedflag = true;
			setProjectionMatrix(zoom);
			break;
	}
}
//<<<<<<<<<<<<<<<<<<<<<<<< myMouse >>>>>>>>>>>>>>>>>
void myMouse(int button, int state, int mx, int my)
{
/* Gets called when a button is clicked.  The arguments are:
button is GLUT_LEFT_BUTTON or GLUT_RIGHT_BUTTON,
state is GLUT_DOWN or GLUT_UP, 
and mx and my are window pixel coordinates, origin at window upper left. */
}

//<<<<<<<<<<<<<<<<<<<<<<<< myLighting >>>>>>>>>>>>>>>>>
void myLighting(bool onoff)
{
	if (onoff)
		::glEnable(GL_LIGHTING);
	else
	{
		::glDisable(GL_LIGHTING);
		return;
	}

//1 set the material properties for objects.
	GLfloat materialSpecular[4] =
		// { 0.1f, 0.1f, 0.1f, 1.0f };
			//Look ok, but a bit dull with materialShininess 2
			//Pops too hard with materialShininess 5
		// { 0.3f, 0.3f, 0.3f, 1.0f }; 
			//Pretty nice flow with materialShininess 2
			//Pops too hard with materialShininess 3
		 //{ 0.45f, 0.45f, 0.45f, 1.0f };
			//Smooth lively flow with materialShininess 2.
		  { 0.6f, 0.6f, 0.6f, 1.0f };
			//Pops  a bit with materialShininess 2.
			//Pops hard glaring white highlights with materialShininess 15.
	GLfloat materialShininess[1] = { 2.0f }; 
		//{ 15.0f } gave too abrupt a glare 
		//was 35.0.  Bigger means smaller sized highlight.  Range is 0.0 to 128.0
	::glMaterialfv( GL_FRONT, GL_SPECULAR, materialSpecular );
	::glMaterialfv( GL_FRONT, GL_SHININESS, materialShininess );
	GLfloat whiteAmbDiff[4]={ 1.0, 1.0, 1.0, 1.0f };
	glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, whiteAmbDiff );
	/* We alter the GL_AMBIENT_AND_DIFFUSE setting with a call to 
	setMaterialColor inside each of our cPolyhedron::draw methods. 
	This helper method makes glMaterialfv calls so that our objects 
	will have intrinsic colors. */

//2 set the LightModel settings, including global ambient lighting.
	GLfloat global_ambient[4] = { 0.75f, 0.75f, 0.75f, 1.0f }; //fairly bright
	::glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
	//::glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );
		/* If I enable GL_LIGHT_MODEL_LOCAL_VIEWER it runs slower but the highlights are more accurate */
	//::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
		/* If I use the GL_LIGHT_MODEL_TWO_SIDE, I have nice light when I'm like inside a teapot. */

//3 set the lights.

	//3.1 Define the lights' properties.
/* The ambient is a directionless light that's simply added in; really since we have
a global_ambient above it's less confusing to have the indivdiual light ambients be zero.
	The diffuse is closest to the "color" of the light.
	The specular is used for color highlights, logically it can match diffuse, but
it looks peppier if its different.
	The position, if it has 0 in the fourth coord, is specifying the (x,y,z) direction
of the light, treated as parallel lines form infinity.  If you want to put the light in the 
position (x,y,z) then you put 1 in the fourth coord, and then the light is viewed as
radiating in all directions form this point. */

/* Let's try and pick four directions of light with roughly one dirction comign in towards
the origin from each vertex of a tetrahedron positioned with a point on the negative z
axis and another point on the yz plane. 
	We'll give different colors for the different directions, for the sake of visual
variety.  But try and make the net sum of light coming down from the positive Z be
white. */
	GLfloat ambient0[] =  { 0.1f, 0.1f, 0.1f, 1.0f }; 
		//Only need to turn ONE light's ambience on,
    GLfloat diffuse0[] =  { 0.1f, 0.1f, 0.4f, 1.0f };//Blue
    GLfloat specular0[] = { 0.1f, 0.1f, 0.4f, 1.0f };
    GLfloat position0[] = { -1.0f, -0.5f, -1.0f, 0.0f }; //Dir to origin from hix, hiy, hiz.

	GLfloat ambient1[] =  { 0.0f, 0.0f, 0.0f, 0.0f };
	GLfloat diffuse1[] =  { 0.4f, 0.1f, 0.1f, 1.0f }; //Red
    GLfloat specular1[] = { 0.4f, 0.1f, 0.1f, 1.0f };
    GLfloat position1[] = { 1.0f, -0.5f, -1.0f, 0.0f }; //Dir to origin from lox, hiy, hiz.
  
	GLfloat ambient2[] =  { 0.0f, 0.0f, 0.0f, 0.0f };
    GLfloat diffuse2[] =  { 0.1f, 0.4f, 0.1f, 1.0f }; //Green
    GLfloat specular2[] = { 0.1f, 0.4f, 0.1f, 1.0f };
    GLfloat position2[] = { 0.0f, 1.0f, -1.0f, 0.0f }; //Dir to origin from zero, loy, hiz
 
	GLfloat ambient3[] =  { 0.0f, 0.0f, 0.0f, 0.0f };
    GLfloat diffuse3[] =  { 0.25f, 0.250f, 0.25f, 1.0f }; //White
    GLfloat specular3[] = { 0.25f, 0.25f, 0.25f, 1.0f };
    GLfloat position3[] = { 0.0f, 0.0f, 1.0f, 0.0f }; //Dir to origin from zero, zero, loz

	/* 3.2 turn off all lights in case you're making this call to change the 
	number of active lights.  There are least eight by default, by the way,
	but we're only using at most four. */
    ::glDisable(GL_LIGHT0);  
    ::glDisable(GL_LIGHT1);  
    ::glDisable(GL_LIGHT2);  
    ::glDisable(GL_LIGHT3);  

	//3.3 Enable your lights
	//Enable Light0
	//Always have at least one light.
    ::glEnable(GL_LIGHT0);  
    ::glLightfv(GL_LIGHT0, GL_AMBIENT, ambient0);
    ::glLightfv(GL_LIGHT0, GL_POSITION, position0);
    ::glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
    ::glLightfv(GL_LIGHT0, GL_SPECULAR, specular0);
	//Enable Light1
	if (NUMBEROFLIGHTS > 1) //NUMBEROFLIGHTS is a #define between 1 and 4.
	{
		::glEnable(GL_LIGHT1);
	    ::glLightfv(GL_LIGHT1, GL_AMBIENT, ambient1);
		::glLightfv(GL_LIGHT1, GL_POSITION, position1);
	    ::glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse1);
		::glLightfv(GL_LIGHT1, GL_SPECULAR, specular1);
	}
	//Enable Light2
	if (NUMBEROFLIGHTS > 2) //NUMBEROFLIGHTS is a #define between 1 and 4.
	{
	    ::glEnable(GL_LIGHT2);  
		::glLightfv(GL_LIGHT2, GL_AMBIENT, ambient2);
		::glLightfv(GL_LIGHT2, GL_POSITION, position2);
		::glLightfv(GL_LIGHT2, GL_DIFFUSE, diffuse2);
		::glLightfv(GL_LIGHT2, GL_SPECULAR, specular2);
	}
	//Enable Light3 
	if (NUMBEROFLIGHTS > 3) //NUMBEROFLIGHTS is a #define between 1 and 4.
	{
		::glEnable(GL_LIGHT3);  
	    ::glLightfv(GL_LIGHT3, GL_AMBIENT, ambient3);
		::glLightfv(GL_LIGHT3, GL_POSITION, position3);
		::glLightfv(GL_LIGHT3, GL_DIFFUSE, diffuse3);
		::glLightfv(GL_LIGHT3, GL_SPECULAR, specular3);
	}
}

//<<<<<<<<<<<<<<<<<<<<<<<< myKeyboard >>>>>>>>>>>>>>>>>
void myMovedMouse(int mx, int my)
{
/*Gets called whenever the mouse moves, and mx and my are the new window
pixel coordinates, origin at window upper left. */
}

//<<<<<<<<<<<<<<<<<<<<<<<< main >>>>>>>>>>>>>>>>>>>>>>
void main(int argc, char** argv)
{
	glutInit(&argc, argv);          // initialize the toolkit
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB ); // set display mode
	glutInitWindowSize(startwindowwidth, int(startwindowwidth/aspect));     // set window size
	glutInitWindowPosition(100, 150); // set window position on screen
	glutCreateWindow("Donutstorm 3, Halloween, 2003, Rudy Rucker"); // open the screen window
	glutDisplayFunc(myDisplay);     // register redraw function
	glutReshapeFunc(myReshape);
	glutKeyboardFunc(myKeyboard);	// register keyboard function.
	glutMouseFunc(myMouse);			// register mouse click function.
	glutMotionFunc(myMovedMouse);   // register mouse move function.
	glutIdleFunc(myAnimate);		// register idle function to call.
	myInit();                   
	glutMainLoop(); 		     // go into a perpetual loop
}
